from __future__ import annotations

import numpy as np
from numpy.typing import ArrayLike
from numpy.random import Generator


def xyz_to_spherical(x_in: float, y_in: float, z_in: float) -> tuple[float, float, float]:
    r = np.linalg.norm([x_in, y_in, z_in])
    theta = np.arccos(z_in)
    phi = np.arctan2(y_in, x_in)

    return r, theta, phi


def spherical_to_xyz(r: float, theta: float, phi: float) -> tuple[float, float, float]:
    sin_theta = np.sin(theta)
    return r*np.cos(phi)*sin_theta, r*np.sin(phi)*sin_theta, r*np.cos(theta)


def get_rotation_matrix(vec_in: np.ndarray, vec_out: np.ndarray):
    rotation = np.eye(3)

    # rotation by angle around given axis
    cos_angle = np.dot(vec_in, vec_out)

    # check for parallel / anti-parallel vectors
    if cos_angle == 1:
        return rotation
    elif cos_angle == -1:
        return -rotation

    else:
        axis = np.cross(vec_in, vec_out)
        scale = np.linalg.norm(axis)
        axis /= scale
        sin_angle = scale / np.linalg.norm(vec_in) / np.linalg.norm(vec_out)

        v_cross = np.array([
            [0,       -axis[2], axis[1]],
            [axis[2], 0,        -axis[0]],
            [-axis[1], axis[0], 0],
        ])

        rotation += sin_angle * v_cross
        rotation += (1-cos_angle) * v_cross @ v_cross

        return rotation


def omega_q(delta: float, eta: float, theta: ArrayLike, phi: ArrayLike) -> ArrayLike:
    cos_theta = np.cos(theta)
    sin_theta = np.sin(theta)
    return np.pi * delta * (3 * cos_theta**2 - 1 + eta * sin_theta**2 * np.cos(2*phi))


def draw_orientation(rng: Generator, size: int | None = None) -> tuple[ArrayLike, ArrayLike]:
    if size is not None:
        z_theta, z_phi = rng.random((2, size))
    else:
        z_theta, z_phi = rng.random(2)
    theta = np.arccos(1 - 2 * z_theta)
    phi = 2 * np.pi * z_phi

    return theta, phi
